/**
 *
 * Project: Crypto, File: RSA.java
 *
 * Copyright 2019 Locusworks LLC.
 * All rights reserved. Federal copyright law prohibits unauthorized reproduction by
 * any means and imposes fines up to $25,000 for violation.  No part of this material
 * may be reproduced, transmitted, transcribed, stored in a retrieval system, copied,
 * modified, duplicated, adapted or translated into another program language in any
 * form or by any means, electronic, mechanical, photocopying, recording, or
 * otherwise, without the prior written permission from Locusworks. Locusworks
 * affirms that Crypto(R) software and data is subject to United States
 * Government Purpose Rights.  Contact Locusworks, 1313 Lawnview Drive
 * Forney TX 75126, (802) 488-0438, for commercial licensing opportunities.
 *
 * IN NO EVENT SHALL LOCUSWORKS BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF LOCUSWORKS HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  NO RESPONSIBILITY IS ASSUMED BY
 * LOCUSWORKS FOR ITS USE, OR FOR ANY INFRINGEMENTS OF PATENTS OR OTHER RIGHTS OF
 * THIRD PARTIES RESULTING FROM ITS USE.  LOCUSWORKS SPECIFICALLY DISCLAIMS ANY
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE AND
 * ACCOMPANYING DOCUMENTATION, IF ANY, PROVIDED HEREUNDER IS PROVIDED "AS
 * IS". LOCUSWORKS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 * ENHANCEMENTS, OR MODIFICATIONS.
 */
package net.locusworks.crypto;

import java.io.ByteArrayInputStream;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.security.Key;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.SecureRandom;
import java.security.interfaces.RSAKey;
import java.util.Base64;

import javax.crypto.Cipher;
import javax.crypto.CipherInputStream;
import javax.crypto.IllegalBlockSizeException;

import org.apache.commons.io.IOUtils;

import net.locusworks.crypto.KeyFile.EncryptionType;

public class RSA {
  
  private static final String ENCRYPTION_TYPE = "RSA";
  private static final String ENCRYPTION_ALGORITHM = "RSA/ECB/PKCS10PADDING";
  private static final String PROVIDER = "SunJCE";
  private static final String RANDOM_ALGORITHM = "SHA1PRNG";
  
  private static final int PADDING_LENGTH = 11;
  private static final int DEFAULT_KEY_LENGTH = 2048;
  
  public static KeyPair generateKeyPair() {
    return generateKeyPair(DEFAULT_KEY_LENGTH);
  }
  
  public static KeyPair generateKeyPair(int keyLength) {
    try {
      KeyPairGenerator kpg = KeyPairGenerator.getInstance(ENCRYPTION_TYPE);
      SecureRandom sr = SecureRandom.getInstance(RANDOM_ALGORITHM);
      kpg.initialize(keyLength, sr);
      return kpg.genKeyPair();
    } catch (Exception ex) {
      throw new RuntimeException(ex);
    }
  }
  
  public static KeyPair loadPrivateKey(String privateKeyFileName) {
    return loadKeyPair(null, privateKeyFileName);
  }
  
  public static KeyPair loadPublicKey(String publicKeyFileName) {
    return loadKeyPair(publicKeyFileName, null);
  }
  
  public static KeyPair loadKeyPair(String publicKey, String privateKey) {
    KeyFile pubKey = KeyFile.read(publicKey);
    KeyFile prvKey = KeyFile.read(privateKey);
    return new KeyPair((PublicKey)pubKey.getKey(), (PrivateKey)prvKey.getKey());
  }
  
  public static boolean generateAndWriteSSHKeys() {
    KeyPair kp = generateKeyPair();
    return writePrivateKey(kp) && writePublicKey(kp, true);
  }
  
  public static boolean generateAndWriteKeyPair() {
    return generateAndWriteKeyPair(DEFAULT_KEY_LENGTH);
  }
  
  public static boolean generateAndWriteKeyPair(String keyPairName) {
    return generateAndWriteKeyPair(keyPairName, DEFAULT_KEY_LENGTH);
  }
  
  public static boolean generateAndWriteKeyPair(int keyLength) {
    return generateAndWriteKeyPair("id_rsa", keyLength);
  }
  
  public static boolean generateAndWriteKeyPair(String keyPairName, int keyLength) {
    KeyPair kp = generateKeyPair(keyLength);
    return writePrivateKey(kp, keyPairName, "PRIVATE KEY") && writePublicKey(kp, keyPairName + ".pub", "PUBLIC KEY");
  }
  
  public static boolean writePrivateKey(KeyPair kp) {
    return writePrivateKey(kp, "id_rsa", "PRIVATE KEY");
  }
  
  public static boolean writePrivateKey(KeyPair kp, String fileName, String description) {
    return writePemFile(kp.getPrivate(), fileName, description);
  }
  
  public static boolean writePublicKey(KeyPair kp) {
    return writePublicKey(kp, false);
  }
  
  public static boolean writePublicKey(KeyPair kp, boolean sshFormat) {
    return writePublicKey(kp, "id_rsa.pub", "PUBLIC KEY", sshFormat);
  }
  
  public static boolean writePublicKey(KeyPair kp, String fileName, String description) {
    return writePemFile(kp.getPublic(), fileName, description, false);
  }
  
  public static boolean writePublicKey(KeyPair kp, String fileName, String description, boolean sshFormat) {
    return writePemFile(kp.getPublic(), fileName, description, sshFormat);
  }
  
  public static int calculateRequiredKeyLength(String message) {
    return (message.getBytes(StandardCharsets.UTF_8).length + PADDING_LENGTH) * 8;
  }
  
  public static String encrypt(Key key, String message) {
    try {
      calculateKeyLength(key, message);
      Cipher cipher = Cipher.getInstance(ENCRYPTION_ALGORITHM, PROVIDER);
      cipher.init(Cipher.ENCRYPT_MODE, key);
      CipherInputStream cis = new CipherInputStream(new ByteArrayInputStream(message.getBytes(StandardCharsets.UTF_8)), cipher);
      byte[] encrypted = IOUtils.toByteArray(cis);
      return Base64.getEncoder().encodeToString(encrypted);
    } catch (Exception ex) {
      throw new RuntimeException(ex);
    }
  }
  
  public static String decrypt(Key key, String message) {
    try {
      Cipher cipher = Cipher.getInstance(ENCRYPTION_ALGORITHM, PROVIDER);
      cipher.init(Cipher.DECRYPT_MODE, key);
      byte[] decoded = Base64.getDecoder().decode(message);
      byte[] plainTextArray = cipher.doFinal(decoded);
      return new String(plainTextArray, StandardCharsets.UTF_8);
    } catch (Exception ex) {
      throw new RuntimeException(ex);
    }
  }

  private static boolean writePemFile(Key key, String fileName, String description) {
    return writePemFile(key, fileName, description, false);
  }
  
  private static boolean writePemFile(Key key, String fileName, String description, boolean sshFormat) {
    try(KeyFile kf = sshFormat ? new KeyFile(key, description, EncryptionType.SSH) : new KeyFile(key, description)) {
      kf.write(fileName);
    }
    return Files.exists(Paths.get(fileName));
  }
  
  private static void calculateKeyLength(Key key, String message) throws IllegalBlockSizeException {
    int keyLength = ((RSAKey)key).getModulus().bitLength();
    int requiredKeyLength = calculateRequiredKeyLength(message);
    if (keyLength < requiredKeyLength) {
      throw new IllegalBlockSizeException(String.format("RSA key size of %d is not large enough to encrypt message of length %d. "
          + "Increase key size to a minimum of %d and re-encrypt with new key", keyLength, message.length(), requiredKeyLength));
    }
  }
}
